#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Author: Saifeddine ALOUI
Description:
Grapph object
"""
import enum
from QGraphViz.DotParser.Node import Node

class GraphType(enum.Enum):
    SimpleGraph=0
    DirectedGraph=1        

class Graph(Node):
    """
    The graph object made of nodes, edges and subgraphs 
    """
    def __init__(self, name, graph_type = GraphType.SimpleGraph, parent_graph=None,  **kwargs):
        Node.__init__(self, name, parent_graph, **kwargs)
        self.parent_graph = parent_graph
        self.current_x=0
        self.current_y=0
        self.nodes=[]
        self.edges=[]
        self.graph_type = graph_type

    def addNode(self, node):
        """
        Adds a node to the graph
        :param node: Node to add to the graph
        """
        self.nodes.append(node)

    def addEdge(self, edge):
        """
        Adds an edge to the graph
        :param edge: An edge to be added to the graph
        """
        self.nodes.append(edge)
   
    def getNodeByName(self, name):
        nodenames = [n.name for n in self.nodes]
        if name in nodenames:
            return self.nodes[nodenames.index(name)]
        else:
            return None




#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Author: Saifeddine ALOUI
Description:
Dot perser implementation
"""

class Edge():
    """
    Describes edges that connect nodes
    """
    def __init__(self, source, dest):
        self.source = source
        self.dest = dest

        source.out_edges.append(self)
        dest.in_edges.append(self)

    



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Author: Saifeddine ALOUI
Description:
Main Class to QGraphViz tool
"""
from PyQt5.QtWidgets import QApplication, QWidget
from PyQt5.QtGui import QPainter, QPen, QBrush, QColor
from PyQt5.QtCore import Qt 
import sys
import enum
import datetime
from QGraphViz.DotParser import DotParser, Node, Edge, Graph, GraphType
import math

class QGraphVizManipulationMode(enum.Enum):
    Static=0
    Nodes_Move_Mode=1
    Edges_Connect_Mode=2
    Node_remove_Mode=3
    Edge_remove_Mode=4
    Subgraph_remove_Mode=5

class QGraphViz(QWidget):
    """
    Main graphviz widget to draw and interact with graphs
    :param parent: A QWidget parent of the QGraphViz widget
    :param engine: The graph processing engine (exemple Dot engine)
    """
    def __init__(
                    self, 
                    parent=None, 
                    engine=None, 
                    show_subgraphs = True,
                    manipulation_mode=QGraphVizManipulationMode.Nodes_Move_Mode,
                    new_edge_created_callback=None, # A callbakc called when a new connection is created between two nodes using the GUI
                    node_selected_callback=None, # A callback called when a node is clicked
                    edge_selected_callback=None, # A callback called when an edge is clicked
                    node_invoked_callback=None, # A callback called when a node is double clicked
                    edge_invoked_callback=None, # A callback called when an edge is double clicked
                ):
        QWidget.__init__(self,parent)
        self.parser = DotParser()
        self.engine=engine
        self.qnodes=[]
        self.qedges=[]
        # Nodes manipulation
        self.manipulation_mode = manipulation_mode
        self.selected_Node = None  
        self.current_pos = [0,0]
        self.mouse_down=False
        self.min_cursor_edge_dist=3
        self.show_subgraphs=show_subgraphs

        self.new_edge_created_callback = new_edge_created_callback
        self.node_selected_callback = node_selected_callback
        self.edge_selected_callback = edge_selected_callback
        self.node_invoked_callback = node_invoked_callback
        self.edge_invoked_callback = edge_invoked_callback

    def build(self):
        self.engine.build()
        """
        for node in self.engine.graph.nodes:
            qnode = QNode(node, self)
            qnode.setParent(self)
            self.qnodes.append(qnode)
        for edge in self.engine.graph.edges:
            qedge = QEdge(edge, self)
            qedge.setParent(self)
            self.qedges.append(qedge)
        """
    def paintSubgraph(self, subgraph, painter, pen, brush):
        if("color" in subgraph.kwargs.keys()):
            pen.setColor(QColor(subgraph.kwargs["color"]))
        else:
            pen.setColor(QColor("black"))

        painter.setPen(pen)
        painter.setBrush(brush)
        gpos = subgraph.global_pos

        painter.drawRect(
                    gpos[0]-subgraph.size[0]/2,
                    gpos[1]-subgraph.size[1]/2,
                    subgraph.size[0], subgraph.size[1])

        if("label" in subgraph.kwargs.keys()):
            painter.drawText(
                gpos[0]-subgraph.size[0]/2,
                gpos[1]-subgraph.size[1]/2,
                subgraph.size[0], subgraph.size[1],
                Qt.AlignCenter|Qt.AlignTop,subgraph.kwargs["label"])

    def paintGraph(self, graph, painter):
        brush = QBrush(Qt.SolidPattern)
        pen=QPen()
        brush.setColor(Qt.white)

        if(self.show_subgraphs):
            for node in graph.nodes:
                if type(node)==Graph:
                    subgraph = node
                    self.paintSubgraph(subgraph, painter, pen, brush)

        for edge in graph.edges:
            if("color" in edge.kwargs.keys()):
                pen.setColor(QColor(edge.kwargs["color"]))
            else:
                pen.setColor(QColor("black"))

            if("width" in edge.kwargs.keys()):
                pen.setWidth(int(edge.kwargs["width"]))
            else:
                pen.setWidth(1)

            painter.setPen(pen)
            painter.setBrush(brush)
            if(edge.source.parent_graph !=graph and not self.show_subgraphs):
                gspos = edge.source.parent_graph.global_pos
            else:
                gspos = edge.source.global_pos

            if(edge.dest.parent_graph !=graph and not self.show_subgraphs):
                gspos = edge.dest.parent_graph.global_pos
            else:
                gdpos = edge.dest.global_pos
            painter.drawLine(gspos[0],gspos[1],
            gdpos[0],
            gdpos[1])

         # TODO : add more painting parameters
        for node in graph.nodes:
            if type(node)!=Graph:
                if("color" in node.kwargs.keys()):
                    pen.setColor(QColor(node.kwargs["color"]))
                else:
                    pen.setColor(QColor("black"))
                gpos = node.global_pos

                painter.setPen(pen)
                painter.setBrush(brush)
                if("shape" in node.kwargs.keys()):
                    if(node.kwargs["shape"]=="box"):
                        painter.drawRect(
                                    gpos[0]-node.size[0]/2,
                                    gpos[1]-node.size[1]/2,
                                    node.size[0], node.size[1])

                    if(node.kwargs["shape"]=="circle"):
                        painter.drawEllipse(
                                    gpos[0]-node.size[0]/2,
                                    gpos[1]-node.size[1]/2,
                                    node.size[0], node.size[1])

                else:
                    painter.drawEllipse(
                                gpos[0]-node.size[0]/2,
                                gpos[1]-node.size[1]/2,
                                node.size[0], node.size[1])


                if("label" in node.kwargs.keys()):
                    painter.drawText(
                        gpos[0]-node.size[0]/2,
                        gpos[1]-node.size[1]/2,
                        node.size[0], node.size[1],
                        Qt.AlignCenter|Qt.AlignTop,node.kwargs["label"])
            else:
                if(self.show_subgraphs):
                    self.paintGraph(subgraph, painter)
                else:
                    subgraph = node
                    self.paintSubgraph(subgraph, painter, pen, brush)

    def paintEvent(self, event):
        painter = QPainter(self) 
        painter.setFont(self.engine.font)
        self.paintGraph(self.engine.graph,painter)
        if( self.manipulation_mode==QGraphVizManipulationMode.Edges_Connect_Mode and 
            self.mouse_down and 
            self.selected_Node is not None):
            bkp = painter.pen()
            pen=QPen(Qt.DashLine)
            painter.setPen(pen)
            painter.drawLine(self.selected_Node.pos[0], self.selected_Node.pos[1],
                             self.current_pos[0],self.current_pos[1])
            painter.setPen(bkp)
        painter.end()

    def new(self, engine):
        """
        Creates a new engine
        :param engine: An engine object (for example a Dot engine)
        """
        self.engine=engine

    def addNode(self, graph, node_name, **kwargs):
        """
        Adds a node to a graph or subgraph
        """
        node = Node(node_name, graph, **kwargs)
        graph.nodes.append(node)
        return node

    def addEdge(self, source, dest, kwargs):
        """
        Connects two nodes from the same subgraph or 
        from two different subgraphs
        If source and dest nodes belong to the same
        Subgraph, the connection added to the subgraph
        if the connection is between different subgraph notes
        the connection is added to the main subgraph 
        """
        edge = Edge(source, dest)
        edge.kwargs=kwargs
        if(source.parent_graph == dest.parent_graph):
            source.parent_graph.edges.append(edge)
        else:
            self.engine.graph.edges.append(edge)

    def addSubgraph(self, parent_graph, subgraph_name, subgraph_type= GraphType.SimpleGraph, **kwargs):
        subgraph = Graph(subgraph_name,subgraph_type, parent_graph, **kwargs)
        subgraph.name = subgraph_name
        subgraph.parent_graph = parent_graph
        parent_graph.nodes.append(subgraph)
        return subgraph

    def removeNode(self, node):
        graph = node.parent_graph
        if(node in graph.nodes):
            idx = graph.nodes.index(node)
            node = graph.nodes[idx]
            for edge in node.in_edges:
                del edge.source.out_edges[edge.source.out_edges.index(edge)]
                if edge.source.parent_graph == edge.dest.parent_graph:
                    del edge.source.parent_graph.edges[edge.source.parent_graph.edges.index(edge)]
                else:
                    del self.engine.graph.edges[self.engine.graph.edges.index(edge)]

            for edge in node.out_edges:
                del edge.source.out_edges[edge.source.out_edges.index(edge)]
                if edge.source.parent_graph == edge.dest.parent_graph:
                    del edge.source.parent_graph.edges[edge.source.parent_graph.edges.index(edge)]
                else:
                    del self.engine.graph.edges[self.engine.graph.edges.index(edge)]
            del graph.nodes[idx]
            self.repaint()

    def removeSubgraph(self, subgraph):
        graph = subgraph.parent_graph
        if(subgraph in graph.subgraphs):
            idx = graph.subgraphs.index(subgraph)
            subgraph = graph.subgraphs[idx]
            del graph.subgraphs[idx]
            self.repaint()

    def removeEdge(self, edge):
        if(edge in self.engine.graph.edges):
            source = edge.source
            dest = edge.dest

            idx = source.out_edges.index(edge)
            del source.out_edges[idx]

            idx = dest.in_edges.index(edge)
            del dest.in_edges[idx]

            if edge.source.parent_graph == edge.dest.parent_graph:
                del edge.source.parent_graph.edges[edge.source.parent_graph.edges.index(edge)]
            else:
                del self.engine.graph.edges[self.engine.graph.edges.index(edge)]

            self.repaint()

    def findSubNode(self, x, y):
        for node in self.engine.graph.nodes:
            gpos=node.global_pos
            if(
                type(node)==Graph and
                gpos[0]-node.size[0]/2<x and gpos[0]+node.size[0]/2>x and
                gpos[1]-node.size[1]/2<y and gpos[1]+node.size[1]/2>y

            ):
                return node
        return None

    def findNode(self, x, y):
        for n in self.engine.graph.nodes:
            gpos=n.global_pos
            if(
                gpos[0]-n.size[0]/2<x and gpos[0]+n.size[0]/2>x and
                gpos[1]-n.size[1]/2<y and gpos[1]+n.size[1]/2>y
            ):
                return n
        return None

    def findEdge(self, x, y):
        for e in self.engine.graph.edges:
            sx=e.source.pos[0] if e.source.pos[0]< e.dest.pos[0] else e.dest.pos[0]
            sy=e.source.pos[1] if e.source.pos[1]< e.dest.pos[1] else e.dest.pos[1]

            ex=e.source.pos[0] if e.source.pos[0]> e.dest.pos[0] else e.dest.pos[0]
            ey=e.source.pos[1] if e.source.pos[1]> e.dest.pos[1] else e.dest.pos[1]

            if(x>sx-self.min_cursor_edge_dist and x<ex+self.min_cursor_edge_dist and
               y>sy-self.min_cursor_edge_dist and y<ey+self.min_cursor_edge_dist):
                x2 = x-sx 
                y2 = y-sy 
                dx = (ex-sx)
                dy = (ey-sy)
                if(dx == 0):
                    if(abs(x2)<self.min_cursor_edge_dist):
                        return e
                elif(dy == 0):
                    if(abs(y2)<self.min_cursor_edge_dist):
                        return e
                else:
                    a = -dy/dx
                    if(abs(a*x2+y2)/math.sqrt(a**2)<self.min_cursor_edge_dist):
                        return e

        return None
    def load_file(self, filename):
        self.engine.graph = self.parser.parseFile(filename)
        self.build()
        self.update()

    def save(self, filename):
        self.parser.save(filename, self.engine.graph)


    def mouseDoubleClickEvent(self, event):
        x = event.x()
        y = event.y()
        n = self.findNode(x,y)
        if n is not None:
            if(self.node_invoked_callback is not None):
                self.node_invoked_callback(n)
        else:
            e = self.findEdge(x, y)
            if e is not None:
                if(self.edge_invoked_callback is not None):
                    self.edge_invoked_callback(e)

        QWidget.mouseDoubleClickEvent(self, event)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            x = event.x()
            y = event.y()
            self.current_pos = [x,y]
            self.mouse_down=True
            n = self.findNode(x,y)
            self.selected_Node = n                
            if(n is None):
                n = self.findSubNode(x,y)
                self.selected_Node = n                

        QWidget.mousePressEvent(self, event)


    def mouseMoveEvent(self, event):
        if self.selected_Node is not None and self.mouse_down:
            x = event.x()
            y = event.y()
            if(self.manipulation_mode==QGraphVizManipulationMode.Nodes_Move_Mode):
                self.selected_Node.pos[0] += x-self.current_pos[0]
                self.selected_Node.pos[1] += y-self.current_pos[1]

            self.current_pos = [x,y]
            self.repaint()
        QWidget.mouseMoveEvent(self, event)

    def mouseReleaseEvent(self, event):
        x = event.x()
        y = event.y()
        n = self.findNode(x,y)   
        if n is None:
            s = self.findSubNode(x,y)     
        if n is None:
            e = self.findEdge(x,y)        
        else:
            e = None
        # Manipulating nodes
        if(self.manipulation_mode==QGraphVizManipulationMode.Nodes_Move_Mode):
            if self.selected_Node is not None and self.mouse_down:
                s = self.findSubNode(x,y)
                if(s is not None and s!=self.selected_Node):
                    if(type(self.selected_Node)==Node):
                        del self.selected_Node.parent_graph.nodes[self.selected_Node.parent_graph.nodes.index(self.selected_Node)]
                        s.nodes.append(self.selected_Node)
                        self.selected_Node.parent_graph = s
                        self.build()
                        self.repaint()
                    if(type(self.selected_Node)==Graph):
                        del self.selected_Node.parent_graph.nodes[self.selected_Node.parent_graph.nodes.index(self.selected_Node)]
                        s.nodes.append(self.selected_Node)
                        self.selected_Node.parent_graph = s
                        self.build()
                        self.repaint()

        # Connecting edges
        if(self.manipulation_mode==QGraphVizManipulationMode.Edges_Connect_Mode):
            if self.selected_Node is not None and self.mouse_down:
                d = n if n is not None else s
                if(d!=self.selected_Node and d is not None):
                    add_the_edge=True
                    if(self.new_edge_created_callback is not None):
                        add_the_edge, kwargs=self.new_edge_created_callback(self.selected_Node, d)
                    else:
                        kwargs={}
                    if add_the_edge:
                        self.addEdge(self.selected_Node, d, kwargs)
                        self.build()
                self.selected_Node=None
        # Removing node
        elif(self.manipulation_mode==QGraphVizManipulationMode.Node_remove_Mode):
            if(n is not None):
                self.removeNode(n)
                self.build()
                self.repaint()

        #Removing edge
        elif(self.manipulation_mode==QGraphVizManipulationMode.Edge_remove_Mode):
            if(e is not None):
                self.removeEdge(e)
                self.build()
                self.repaint()
        # Remiving Subgraph
        elif(self.manipulation_mode==QGraphVizManipulationMode.Subgraph_remove_Mode):
            if(s is not None):
                self.removeSubgraph(s)
                self.build()
                self.repaint()

        # Inform application
        if(n is not None):
            if(self.node_selected_callback is not None):
                self.node_selected_callback(n)

        if( e is not None):
            if(self.edge_selected_callback is not None):
                self.edge_selected_callback(e)

        QWidget.mouseReleaseEvent(self, event)
        self.mouse_down=False
        self.repaint()

        



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Author: Saifeddine ALOUI
Description:
Dot perser implementation
"""

class Node():
    """
    The dot graphviz engine
    """
    def __init__(self, name, parent_graph, **kwargs):
        self.name = name
        self.parent_graph = parent_graph
        self.pos=[0,0]
        self.size=[1,1]
        self.kwargs=kwargs
        # Nodes
        self.in_edges=[]
        self.out_edges=[]
        self.processed = 0
        
    @property
    def global_pos(self):
        if(self.parent_graph is not None):
            return [self.pos[0]+self.parent_graph.pos[0], self.pos[1]+self.parent_graph.pos[1]]
        else:
            return self.pos



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Author: Saifeddine ALOUI
Description:
Dot perser implementation
"""
from QGraphViz.DotParser.Graph import Graph, GraphType
from QGraphViz.DotParser.Node import Node
from QGraphViz.DotParser.Edge import Edge

from itertools import groupby
class DotParser():
    """
    The dot language parser
    """
    def __init__(self):
        pass

    def isParam_line(self, data):
        """
        Detect if the following is a parameters line
        """
        l=len(data)
        for (i,c) in enumerate(data):
            if c=="=":
                return True
            if c=="[":
                return False
            if i < l-1:
                if data[i:i+2]=="--":
                    return False
                if data[i:i+2]=="->":
                    return False
        return False  

    def isNode_line(self, data):  
        """
        Detects if this is a node line
        """
        l=len(data)
        for (i,c) in enumerate(data):
            if c=="[":
                return True
            if c=="}":
                return False
            if i < l-1:
                if data[i:i+2]=="--":
                    return False
                if data[i:i+2]=="->":
                    return False
        return True

    def isEdge_line(self, data, connection_sign="--"): 
        """
        Detects if this is an edge line
        """ 
        l=len(data)
        for (i,_) in enumerate(data):
            if i < l-1:
                if data[i:i+2]==connection_sign:
                    return True
        return False

    def has_params(self, data):
        """
        Detects if a node or cluster has parameters
        """
        for c in data:
            if c=="[":
                return True
            if c=="{":
                return False
        return False

    def find_end_brackets(self, data):
        """
        Suppose data starts after a bracket, find the index of the closing bracket
        """
        dquotes_level = 0
        quotes_level = 0
        bracket_level = 0
        braces_level = 0
        for i, c in enumerate(data):
            if c == "}" and braces_level == 0 and bracket_level == 0 and quotes_level == 0 and dquotes_level == 0:
                return i
            else:
                if c == "\"":
                    if dquotes_level==0:
                        dquotes_level += 1
                    else:
                        dquotes_level = 0
                if c == "'":
                    if quotes_level==0:
                        quotes_level += 1
                    else:
                        quotes_level = 0
                if c == "{":
                    bracket_level += 1
                elif c == "}":
                    bracket_level -= 1

                if c == "[":
                    braces_level += 1
                elif c == "]":
                    braces_level -= 1
        return None

    def split(self, s):
        """
        Splits a string while respecting ponctuation
        """
        parts = []
        dquotes_level = 0
        quotes_level = 0
        bracket_level = 0
        braces_level = 0
        current = []
        # trick to remove special-case of trailing chars
        for c in (s + ","):
            if c == "," and braces_level == 0 and bracket_level == 0 and quotes_level == 0 and dquotes_level == 0:
                parts.append("".join(current))
                current = []
            else:
                if c == "\"":
                    if dquotes_level==0:
                        dquotes_level += 1
                    else:
                        dquotes_level = 0
                if c == "'":
                    if quotes_level==0:
                        quotes_level += 1
                    else:
                        quotes_level = 0
                if c == "{":
                    bracket_level += 1
                elif c == "}":
                    bracket_level -= 1

                if c == "[":
                    braces_level += 1
                elif c == "]":
                    braces_level -= 1
                current.append(c)
        return parts



    def find_params(self, data):
        """
        Finds parameters
        """
        try:
            start_idx=data.index("[")
            end_index=data.rindex("]")

            strparams = self.split(data[start_idx+1:end_index])#.split(",")
            params={}
            for param in strparams:
                vals = param.split("=")
                params[vals[0].strip()] = "=".join(vals[1:]).strip()
            return params, start_idx, end_index
        except:
            return None, 0, 0


    def parse_Graph(self, data, graph, connection_sign="--"):
        """
        Parses a graph (or a subgraph)
        """
        while len(data)>0:
            lines = data.split("\n",1)
            line = lines[0].strip()
            if(len(lines)>=2):
                data=lines[1]
            else:
                data=""
            if len(line)==0:
                continue
            if "subgraph" in line:
                single_line=False
                if "{" in line:
                    start_brackets_pos = line.index("{")
                    name = line[9:start_brackets_pos]
                    if "cluster_" in name:
                        name = name.replace("cluster_","")
                    end_brackets_pos = self.find_end_brackets(line[start_brackets_pos+1:])
                    if (end_brackets_pos is not None):
                        single_line=True
                else:
                    name = line[9:].strip()
                subgraph = Graph(name, graph.graph_type, graph)
                graph.nodes.append(subgraph)
                if single_line:
                    self.parse_Graph(lines[0][start_brackets_pos+1:end_brackets_pos], subgraph, connection_sign)
                else:
                    end_brackets_pos = self.find_end_brackets(lines[1][start_brackets_pos+1:])
                    self.parse_Graph(lines[1][:end_brackets_pos], subgraph, connection_sign)
            else:
                if self.isParam_line(lines[0]):
                    title,val = lines[0].split("=",1)
                    graph.kwargs[title]=val
                    
                elif self.isNode_line(lines[0]):
                    params, params_start_idx, params_end_index =self.find_params(lines[0])
                    if(params is not None):
                        subname = lines[0][0:params_start_idx].strip()
                    else:
                        subname = lines[0].strip()
                    node = Node(subname, graph)
                    if(params is not None):
                        node.kwargs = params
                    graph.nodes.append(node)
                elif self.isEdge_line(lines[0], connection_sign):
                    link_idx=lines[0].index(connection_sign)
                    # find parameters
                    source_node_name = lines[0][:link_idx].strip()
                    params, params_start_idx, params_end_index =self.find_params(lines[0])
                    
                    if(params is not None):
                        dest_node_name = lines[0][link_idx+3:params_start_idx].strip()
                    else:
                        dest_node_name = lines[0][link_idx+3:].strip()

                    source_node  = graph.getNodeByName(source_node_name)
                    dest_node  = graph.getNodeByName(dest_node_name)
                    if(source_node is not None and dest_node is not None):
                        edge = Edge(source_node, dest_node)
                        if(params is None):
                            edge.kwargs={}
                        else:
                            edge.kwargs=params
                        graph.edges.append(edge)


    def parseFile(self, filename):
        """
        Parses a file
        """
        graph=Graph("main_graph")
        with open(filename,"r") as fi:
            data=fi.read()
            try:
                graph_idx= data.index("dgraph")
                connection_sign = "->"
                print("found dgraph at {}".format(graph_idx))
                graph=Graph("main_graph",GraphType.DirectedGraph)
            except:
                try:
                    graph_idx= data.index("graph")
                    connection_sign = "--"
                    print("found dgraph at {}".format(graph_idx))
                    graph=Graph("main_graph")
                except:
                    raise Exception("Syntax error")
            data=data[graph_idx+5:]
            start_id=data.index("{")
            end_id=data.rindex("}")
            self.parse_Graph(data[start_id+1:end_id], graph, connection_sign)
        return graph

    def populate_file(self, graph, connection_sign, fi):
        """
        Fills a gv file with data from the graph
        """
        for node in graph.nodes:
            if(type(node)==Graph):
                fi.write("    subgraph cluster_{}".format(node.name))
                fi.write("{\n")
                fi.write("\n".join(["{}={}".format(k,v) for k,v in node.kwargs.items()]))
                fi.write("\n")
                self.populate_file(node, connection_sign, fi)
                fi.write("    }\n")
            else:
                fi.write("    {} [{}]\n".format(node.name, ",".join(["{}={}".format(k,v) for k,v in node.kwargs.items()])))
        for edge in graph.edges:
            if(not edge.kwargs):
                fi.write("    {} {} {}\n".format(edge.source.name, connection_sign, edge.dest.name))
            else:
                fi.write("    {} {} {} [{}]\n".format(edge.source.name, connection_sign, edge.dest.name, ",".join(["{}={}".format(k,v) for k,v in edge.kwargs.items()])))

    def save(self, filename, graph):
        """
        Saves a graph to a gv file
        """
        with open(filename,"w") as fi:
            if graph.graph_type == GraphType.SimpleGraph:
                fi.write("graph {\n")
                self.populate_file(graph, "--", fi)
                fi.write("}")
            if graph.graph_type == GraphType.DirectedGraph:
                fi.write("dgraph {\n")
                self.populate_file(graph, "->", fi)
                fi.write("}")
